from ngsolve import *
import netgen.geom2d as geom2d
from netgen.geom2d import *
from math import pi
import numpy as np
import os

# Solve Helm(u):= u - Lap(u) = f in Y:=(0,1)^2 with periodic b.c. 
# DG scheme: Find u in Vh s.t. int_Y Helm(u)*Helm(v) + int_faces (mu[grad(u)]*[grad(v)]+eta[u][v]) = int_Y f*Helm(v) for all v in Vh

hdiscretization = 1/64
deg = 3.0 

# Source f and true soln U
f = (1.0+8.0*pi*pi)*cos(2.0*pi*x)*cos(2.0*pi*y)

U = cos(2.0*pi*x)*cos(2.0*pi*y)
Ux = -2.0*pi*sin(2.0*pi*x)*cos(2.0*pi*y)
Uy = -2.0*pi*cos(2.0*pi*x)*sin(2.0*pi*y)
Uxx = -4.0*pi*pi*cos(2.0*pi*x)*cos(2.0*pi*y)
Uxy = 4.0*pi*pi*sin(2.0*pi*x)*sin(2.0*pi*y)
Uyy = -4.0*pi*pi*cos(2.0*pi*x)*cos(2.0*pi*y)

# Mesh and FE space
periodic = SplineGeometry()
pnums = [periodic.AppendPoint(*p) for p in [ (0,0), (1,0), (1,1), (0,1) ]]
lright = periodic.Append ( ["line", pnums[0], pnums[1]],bc="periodic")
btop = periodic.Append ( ["line", pnums[1], pnums[2]], bc="periodic")
periodic.Append ( ["line", pnums[3], pnums[2]], leftdomain=0, rightdomain=1, copy=lright, bc="periodic")
periodic.Append ( ["line", pnums[0], pnums[3]], leftdomain=0, rightdomain=1, copy=btop, bc="periodic")
mesh = Mesh(periodic.GenerateMesh(maxh=hdiscretization))

FEspace = L2(mesh, order = deg, dgjumps=True)
h = specialcf.mesh_size

# penalization parameters
mu = (10.0*deg*deg)/h
eta = (10.0*deg*deg*deg*deg)/(h*h*h)

# some macros
def hess(w): return w.Operator("hesse")
def d0(u): return grad(u)[0]
def d1(u): return grad(u)[1]
def dxx(w): return hess(w)[0,0]
def dxy(w): return hess(w)[0,1]
def dyx(w): return hess(w)[0,1]
def dyy(w): return hess(w)[1,1]
def Helm(w): return (w-dxx(w)-dyy(w))

# Variational formulation
u = FEspace.TrialFunction()
v = FEspace.TestFunction()
uh = GridFunction(FEspace,autoupdate=True)
Rhs = LinearForm(FEspace)
a = BilinearForm(FEspace, symmetric=False)

Rhs += (f*Helm(v))*dx(bonus_intorder=10)	
a += (Helm(u)*Helm(v))*dx(bonus_intorder=10) + (mu*(grad(u)-grad(u).Other())*(grad(v)-grad(v).Other())+eta*(u-u.Other())*(v-v.Other()))*dx(skeleton=True)

Rhs.Assemble()
a.Assemble()
uh.vec.data = a.mat.Inverse(FEMspace.FreeDofs())*Rhs.vec

# Compute error to true solution	
jumperr = sqrt(Integrate ( (1.0/h)*(grad(uh)-grad(uh).Other())*(grad(uh)-grad(uh).Other())*dx(element_boundary=True),mesh) + Integrate ( (1.0/(h*h*h))*(uh-uh.Other())*(uh-uh.Other())*dx(element_boundary=True),mesh))
L2err = sqrt(Integrate((uh-U)*(uh-U),mesh))
H1err = sqrt(Integrate((Ux-d0(uh))*(Ux-d0(uh))+(Uy-d1(uh))*(Uy-d1(uh)),mesh))
H2err = sqrt(Integrate((dxx(uh)-Uxx)*(dxx(uh)-Uxx)+(dxy(uh)-Uxy)*(dxy(uh)-Uxy)+(dyx(uh)-Uxy)*(dyx(uh)-Uxy)+(dyy(uh)-Uyy)*(dyy(uh)-Uyy),mesh))

Draw(uh)

print('------------------------ h = ', hdiscretization,'-----------------------------')
print('|U-uh|_jump = ',jumperr)
print('||U-uh||_L2 = ', L2err)
print('|U-uh|_H1 = ', H1err)
print('|U-uh|_H2 = ', H2err)
print('------------------------------------------------------------------')
